//====================================================================
// Name          : Bus Structure Extraction Enhancement Code
// Created       : Kevin Vinding
// Description   : Macros used for extracting meta-information on
//                 bus structures.
//====================================================================
`ifndef BUS_EXTRACTION_SV
`define BUS_EXTRACTION_SV

// Types of information:
`define INFO_READ "ReadBus"
`define INFO_FORCE "ForceValue"
`define INFO_INTERCONNECT "Interconnect"
`define INFO_MARK "MarkBus"
`define INFO_RELAY "Relay"


integer forceValue = 10;

function integer getNewForceValue();
	return forceValue++;
endfunction : getNewForceValue


// -- Writes values to log
`define WRITE_VALUES_TO_LOG(MODULE_NAME, TYPE_OF_INFORMATION, VALUE_1, VALUE_2 = 0) \
`ifdef ENABLE_BUS_EXTRACTION \
	$display("#bus#%s#%s#%h#%h#", MODULE_NAME, TYPE_OF_INFORMATION, VALUE_1, VALUE_2); \
`endif

// -- Writes text (with associated value) to log
`define WRITE_VALUE_AND_TEXT_TO_LOG(MODULE_NAME, TYPE_OF_INFORMATION, VALUE, TEXT) \
`ifdef ENABLE_BUS_EXTRACTION \
	$display("#bus#%s#%s#%h#%s#", MODULE_NAME, TYPE_OF_INFORMATION, VALUE, TEXT); \
`endif

// -- Forces the value VALUE on DATA_BUS. Parameter MODULE_NAME defaults to the hierarchical path of the caller.
`define FORCE_VALUE_ON_BUS(DATA_BUS, VALUE, MODULE_NAME=$sformatf("%m")) \
`ifdef ENABLE_BUS_EXTRACTION \
	force DATA_BUS = VALUE; \
	`WRITE_VALUES_TO_LOG(MODULE_NAME, `INFO_FORCE, VALUE, ) \
`endif

// -- Forces a random value on DATA_BUS. Parameter MODULE_NAME defaults to the hierarchical path of the caller.
`define FORCE_RANDOM_VALUE_ON_BUS(DATA_BUS, MODULE_NAME=$sformatf("%m")) \
`ifdef ENABLE_BUS_EXTRACTION \
	begin \
		static int rand_value = $random(); \
		`FORCE_VALUE_ON_BUS(DATA_BUS, rand_value, MODULE_NAME) \
	end \
`endif

// -- Releases DATA_BUS after a value has been forced by e.g., using FORCE_RANDOM_VALUE_ON_BUS(...).
`define RELEASE_BUS(DATA_BUS) \
`ifdef ENABLE_BUS_EXTRACTION \
	release DATA_BUS; \
`endif

// -- Forces a random value on DATA_BUS, waits 2 delta-cycles, then releases DATA_BUS. 
//    Parameter MODULE_NAME defaults to the hierarchical path of the caller.
`define FORCE_AND_RELEASE_BUS(DATA_BUS, BUS_NAME="", MODULE_NAME=$sformatf("%m")) \
`ifdef ENABLE_BUS_EXTRACTION \
	begin \
		static int value_to_force = getNewForceValue(); \
		`FORCE_VALUE_ON_BUS(DATA_BUS, value_to_force, MODULE_NAME) \
		#0; \
		`WRITE_VALUE_AND_TEXT_TO_LOG(MODULE_NAME, `INFO_READ, DATA_BUS, BUS_NAME) \
		#0; \
		`RELEASE_BUS(DATA_BUS) \
	end \
`endif

// -- Forces random values on all busses in DATA_BUS_ARRAY, and releases them after 2 delta-cycles.
// -- Note: The data-bus array is expected to be on the form: [ARRAY_SIZE-1:0][BUS_SIZE-1:0] DATA_BUS_ARRAY
`define FORCE_AND_RELEASE_BUS_ARRAY(DATA_BUS_ARRAY, ARRAY_SIZE, BUS_SIZE, BUS_NAME="", MODULE_NAME=$sformatf("%m")) \
`ifdef ENABLE_BUS_EXTRACTION \
	begin \
		logic [ARRAY_SIZE-1:0][BUS_SIZE-1:0] ValueForBusArray; \
		foreach (ValueForBusArray[i]) begin \
			ValueForBusArray[i] = getNewForceValue(); \
		end \
		force DATA_BUS_ARRAY = ValueForBusArray; \
		foreach (ValueForBusArray[i]) begin \
			`WRITE_VALUES_TO_LOG(MODULE_NAME, `INFO_FORCE, ValueForBusArray[i], ) \
		end \
		#0; \
		foreach (DATA_BUS_ARRAY[i]) begin \
			`WRITE_VALUE_AND_TEXT_TO_LOG(MODULE_NAME, `INFO_READ, DATA_BUS_ARRAY[i], $sformatf("%s[%0d]", BUS_NAME, i)) \
		end \
		#0; \
		`RELEASE_BUS(DATA_BUS_ARRAY) \
	end \
`endif

// -- NEw read-bus macro. Change name to READ_BUS when appropiate
`define READ_BUS(DATA_BUS, BUS_NAME="", MODULE_NAME=$sformatf("%m")) \
`ifdef ENABLE_BUS_EXTRACTION \
	begin \
		#0; \
		`WRITE_VALUE_AND_TEXT_TO_LOG(MODULE_NAME, `INFO_READ, DATA_BUS, BUS_NAME) \
	end \
`endif

// -- New read-bus-array macro. Change name to READ_BUS_ARRAY when appropiate
`define READ_BUS_ARRAY(DATA_BUSSES, BUS_NAME="", MODULE_NAME=$sformatf("%m")) \
`ifdef ENABLE_BUS_EXTRACTION \
	begin \
		#0; \
		foreach (DATA_BUSSES[i]) begin \
			`WRITE_VALUE_AND_TEXT_TO_LOG(MODULE_NAME, `INFO_READ, DATA_BUSSES[i], $sformatf("%s[%0d]", BUS_NAME, i)) \
		end \
	end \
`endif

// -- Finish the simulation early.
`define FINISH_SIMULATION_EARLY() \
`ifdef ENABLE_BUS_EXTRACTION \
	#1; \
	$display("Simulation ends early due to bus structure extraction."); \
	$finish(); \
`endif

// -- Uses a connection matrix to define which bus in FIRST_BUS_ARRAY array are connected to which
//    bus in SECOND_BUS_ARRAY array (after a delta-cycle).
// -- Note: CONNECTION_MATRIX first dimension must equal that of the FIRST_BUS_ARRAY. Similarly,
//    the second dimension must equal SECOND_BUS_ARRAY.
`define INTERCONNECT_BUSSES(CONNECTION_MATRIX, FIRST_BUS_ARRAY, SECOND_BUS_ARRAY, MODULE_NAME=$sformatf("%m")) \
`ifdef ENABLE_BUS_EXTRACTION \
	begin \
		#0; \
		foreach (FIRST_BUS_ARRAY[i]) begin \
			foreach (SECOND_BUS_ARRAY[j]) begin \
				if (CONNECTION_MATRIX[i][j] != 0) begin \
					`WRITE_VALUES_TO_LOG(MODULE_NAME, `INFO_INTERCONNECT, FIRST_BUS_ARRAY[i], SECOND_BUS_ARRAY[j]) \
				end \
			end \
		end \
	end \
`endif


// -- Defines that DATA_BUS_OUT is an relay-value of DATA_BUS_IN
`define RELAY(DATA_BUS_IN, DATA_BUS_OUT, MODULE_NAME=$sformatf("%m")) \
`ifdef ENABLE_BUS_EXTRACTION \
	begin \
		#0; \
		`WRITE_VALUES_TO_LOG(MODULE_NAME, `INFO_RELAY, DATA_BUS_IN, DATA_BUS_OUT) \
	end \
`endif

// -- Defines that DATA_BUSSES_OUT is relay-values of DATA_BUSSES_IN
//	  Note: DATA_BUSSES_IN must be same size as DATA_BUSSES_OUT
`define RELAY_ARRAY(DATA_BUSSES_IN, DATA_BUSSES_OUT, MODULE_NAME=$sformatf("%m")) \
`ifdef ENABLE_BUS_EXTRACTION \
	begin \
		#0; \
		foreach (DATA_BUSSES_IN[i]) begin \
			`WRITE_VALUES_TO_LOG(MODULE_NAME, `INFO_RELAY, DATA_BUSSES_IN[i], DATA_BUSSES_OUT[i]) \
		end \
	end \
`endif

// -- Mark the DATA_BUS as Master
`define MARK_BUS_AS_MASTER(DATA_BUS, MODULE_NAME=$sformatf("%m")) \
`ifdef ENABLE_BUS_EXTRACTION \
	begin \
		#0; \
		`WRITE_VALUE_AND_TEXT_TO_LOG(MODULE_NAME, `INFO_MARK, DATA_BUS, "Master") \
	end \
`endif

// -- Mark the DATA_BUS as Slave
`define MARK_BUS_AS_SLAVE(DATA_BUS, MODULE_NAME=$sformatf("%m")) \
`ifdef ENABLE_BUS_EXTRACTION \
	begin \
		#0; \
		`WRITE_VALUE_AND_TEXT_TO_LOG(MODULE_NAME, `INFO_MARK, DATA_BUS, "Slave") \
	end \
`endif

// -- Mark several DATA_BUSSES as Master
`define MARK_BUSSES_AS_MASTER(DATA_BUSSES, MODULE_NAME=$sformatf("%m")) \
`ifdef ENABLE_BUS_EXTRACTION \
	begin \
		#0; \
		foreach (DATA_BUSSES[i]) begin \
			`WRITE_VALUE_AND_TEXT_TO_LOG(MODULE_NAME, `INFO_MARK, DATA_BUSSES[i], "Master") \
		end \
	end \
`endif

// -- Mark several DATA_BUSSES as Slave
`define MARK_BUSSES_AS_SLAVE(DATA_BUSSES, MODULE_NAME=$sformatf("%m")) \
`ifdef ENABLE_BUS_EXTRACTION \
	begin \
		#0; \
		foreach (DATA_BUSSES[i]) begin \
			`WRITE_VALUE_AND_TEXT_TO_LOG(MODULE_NAME, `INFO_MARK, DATA_BUSSES[i], "Slave") \
		end \
	end \
`endif

`endif