#####################################################################################################################
# Name:         Bus Info Processor
# Author:       Kevin Vinding
# Description:  Processes and maps out bus information based on datapoints obtained from bus_info_parser.py
#####################################################################################################################
import datapoint as dp

INFO_READ = "ReadBus"
INFO_FORCE = "ForceValue"
INFO_INTERCONNECT = "Interconnect"
INFO_MARK = "MarkBus"
INFO_RELAY = "Relay"
NAME_DELIMITER = '.'

# Get a set of modules from a list of data points
def get_modules_from_data_points(list_data_points):
    modules = set()
    for data_point in list_data_points:
        modules.add(data_point.name)
    return modules


# Reduce the name by only include the last region of the name
def reduce_name(name):
    split_name = name.split(NAME_DELIMITER)
    return split_name[len(split_name)-1]


# Verify that a value is an int and not the value 0
def verify_value(string_value):
    try:
        value_as_int = int(string_value, 16)
        if value_as_int == 0:
            return False
        else:
            return True
    except:    
        return False


# Get a dictionary of busses bewteen modules from a list of data points, with the written value as key
def get_busses_from_data_points(list_data_points):
    busses = dict()
    # Iterate through all data points that are read- or write operations.
    for data_point in list_data_points:
        if data_point.info_type == INFO_READ or data_point.info_type == INFO_FORCE:
            if verify_value(data_point.value_1):

                # Create new entry if not in dict
                if data_point.value_1 not in busses:
                    busses[data_point.value_1] = Bus(data_point.value_1)

                # Add info to entry
                if data_point.info_type == INFO_READ:
                    busses[data_point.value_1].add_read_source(data_point.name)
                    busses[data_point.value_1].add_name(data_point.value_2)
                elif data_point.info_type == INFO_FORCE:
                    busses[data_point.value_1].set_write_source(data_point.name)
    return busses


# Update an existing connecton list with a new connection
def update_connection_list(connection_list, new_connection):
    for connection in connection_list:
        # If it's inside list, discard the new connection
        if connection == new_connection:
            return connection_list
        # If it's the inverted instace, update it to be bidirectional
        if connection.get_inverted() == new_connection:
            connection.bidirectional = True
            return connection_list
    # Else add to the list
    connection_list.append(new_connection)
    return connection_list
            

# Get a list of connections between modules, based on a buss dictionary
def get_connections_from_buss_dict(dict_busses):
    connections = list()
    for bus in dict_busses:
        for read_source in dict_busses[bus].read_sources:
            if read_source != dict_busses[bus].get_write_source():
                new_connection = Connection(dict_busses[bus].get_write_source(), read_source, dict_busses[bus].get_ran_name())
                connections = update_connection_list(connections, new_connection) 
    return connections


# Retrieve a list of masters and which slave is connected to them
def get_master_list(datapoints):
    # 1: get which modules are interconnecting modules
    interconnect_modules = set()
    for data_point in datapoints:
        if data_point.info_type == INFO_INTERCONNECT:
            interconnect_modules.add(data_point.name)

    # 2: get the values on its ports, which are master/slave
    #    and which master interconnect to which slave
    interconnections = list()
    interconnect_ports = set()
    relays = list()
    relay_names = set()
    for data_point in datapoints:
        # Get value and type
        if data_point.info_type == INFO_MARK:
            interconnect_ports.add(Port(data_point.value_1, data_point.value_2))
        # Get interconnections
        if data_point.info_type == INFO_INTERCONNECT:
            interconnections.append(Connection(data_point.value_1, data_point.value_2))
        # Get relays
        if data_point.info_type == INFO_RELAY:
            relays.append(Connection(data_point.value_1, data_point.value_2))
            relay_names.add(data_point.name)

    # 3: Update port values to use relay-values if neccessary
    interconnect_updated = True
    while interconnect_updated:
        interconnect_updated = False
        # Check every interconnecting port up against the relay list
        for p in interconnect_ports:
            for r in relays:
                # On hit, change the value, and change any value in the interconnecting as well
                if p.value == r.source:
                    interconnect_updated = True
                    old_value = p.value
                    p.value = r.target
                    for i in interconnections:
                        if i.target == old_value:
                            i.target = r.target
                        if i.source == old_value:
                            i.source = r.target

    # 4: Get which module reads the value out of interc.
    masters = dict()
    slaves = list()
    for data_point in datapoints:
        if data_point.info_type == INFO_READ and data_point.name not in interconnect_modules and data_point.name not in relay_names:
           for p in interconnect_ports:
                if data_point.value_1 == p.value:
                    if p.port_type == "Slave":
                        slaves.append(Slave(data_point.name, data_point.value_1))
                    elif p.port_type == "Master":
                        if data_point.name not in masters:
                            masters[data_point.name] = Master(data_point.name)
                        masters[data_point.name].master_values.add(data_point.value_1)

    # 5: Collect everything and return
    # Add interconnection values
    for m in masters:
        for i in interconnections:
            if i.source in masters[m].master_values:
                masters[m].slave_values.add(i.target)
            elif i.target in masters[m].master_values:
                masters[m].slave_values.add(i.source)
    # Add slaves
    for m in masters:
        for s in slaves:
            if s.value in masters[m].slave_values:
                masters[m].slaves.add(s.name)

    return masters


class Master:
    name = ""
    master_values = set()
    slave_values = set()
    slaves = set()

    def __init__(self, name):
        self.name = name
        self.master_values = set()
        self.slave_values = set()
        self.slaves = set()

    def __repr__(self):
        return 'Name: {0}\n\tMaster Value: {1}\n\tSlave Values: {2}\n\tSlaves: {3}'.format(
            self.name, self.master_values, self.slave_values, self.slaves
        )

    def __str__(self):
        return 'Name: {0}\n\tMaster Value: {1}\n\tSlave Values: {2}\n\tSlaves: {3}'.format(
            self.name, self.master_values, self.slave_values, self.slaves
        )


class Slave:
    name = ""
    value = ""

    def __init__(self, name, value):
        self.name = name
        self.value = value


class Port:
    value = ""
    port_type = ""

    def __init__(self, value, port_type):
        self.value = value
        self.port_type = port_type


class Bus:
    __value = ""
    __write_source = ""
    names = set()
    read_sources = set()

    def __init__(self, value):
        self.__value = value
        self.__write_source = ""
        self.names = set()
        self.read_sources = set()

    def __repr__(self):
        return 'Name: {0}\n\tValue: {1}\n\tWrite Source: {2}\n\tRead Sources: {3}'.format(
            self.names, self.__value, self.__write_source, self.read_sources
        )

    def __str__(self):
        return 'Name: {0}\n\tValue: {1}\n\tWrite Source: {2}\n\tRead Sources: {3}'.format(
            self.names, self.__value, self.__write_source, self.read_sources
        )

    def add_read_source(self, name_of_source):
        self.read_sources.add(name_of_source)

    def set_write_source(self, name_of_source):
        if self.__write_source == "":
            self.__write_source = name_of_source
            return
        else:
            print('Warning: Tried to overwrite write-source "{}" with "{}" (Bus value: {}).'.format(self.__write_source, name_of_source, self.__value))
            return -1

    def get_write_source(self):
        return self.__write_source

    def get_value(self):
        return self.__value

    def add_name(self, name_of_bus):
        self.names.add(name_of_bus)

    def get_ran_name(self):
        if len(self.names) == 0:
            return "None"
        else:
            ran_name = self.names.pop()
            self.names.add(ran_name)
            return ran_name

class Connection:
    source = ""
    target = ""
    name = ""
    bidirectional = False

    def __init__(self, source, target, name=""):
        self.source = source
        self.target = target
        self.name = name
        self.bidirectional = False

    def __repr__(self):
        return "Source: {0}, Target: {1}, Name:{3}, Is Bidirectional: {2}".format(
            self.source, self.target, self.bidirectional, self.name
        )

    def __str__(self):
        return "Source: {0}, Target: {1},Name:{3}, Is Bidirectional: {2}".format(
            self.source, self.target, self.bidirectional, self.name
        )
    
    def __eq__(self, other):
        if self.source == other.source and self.target == other.target:
            return 1
        else:
            return 0

    def get_inverted(self):
        inv_connection = Connection(self.target, self.source, self.name)
        inv_connection.bidirectional = True
        return inv_connection

