#####################################################################################################################
# Name:         Bus Tool (main file)
# Author:       Kevin Vinding
# Description:  Main file for analysing and visualising the meta-information obtained from extraction code.
#####################################################################################################################
import sys
import getopt
import os
import bus_info_parser as bip
import bus_info_processor as pro
from graphviz import Digraph

CMD_PDF = '/cad/gnu/anaconda/3-2.4.0/bin/dot -Tpdf bus_structure.gv -o graph.pdf && okular graph.pdf'

data_loaded = False
enable_reduced_names = True
enable_debug_mode = False
data_name = 'None'
module_set = set()
busses_dict = dict()
masters = dict()


def add_modules_to_dot(dot, list_of_modules):
    dot.attr('node', shape='box')
    for module in list_of_modules:
        reduced_name = reduce_hierarchical_path_name(module)
        if enable_reduced_names:
            dot.node(module, label=reduced_name)
        else:
            dot.node(module, label=module)
    return dot


def add_connections_to_dot(dot, list_of_connections):
    for con in list_of_connections:
        if con.source != "" and con.target != "":
            if con.bidirectional:
                dot.attr('edge', dir='none', style='solid', penwidth='5') #, penwidth = 2, style='solid')
            else:
                dot.attr('edge', dir='forward', style='solid', penwidth='1') #, penwidth=1, style='dashed')
            dot.edge(con.source, con.target)
    return dot


def reduce_hierarchical_path_name(hierarchical_name):
    separate_hier_name = hierarchical_name.split('.')
    return separate_hier_name[len(separate_hier_name)-1]


# Ask a yes/no-question, then return answer as bool
def y_n_input(question):
    while True:
        answer = input(question).lower()
        if answer == 'y' or answer == 'yes':
            return True
        elif answer == 'n' or answer == 'no':
            return False
        else:
            print('Invalid answer. Use "y" or "n".')
    return


# Remove extension from file name (e.g., .txt or .pdf)
def remove_extension_from_name(name):
    return name.split('.')[0]


# Generates a diagram of the (imported) data as a .pdf file
def generate_connection_diagram():
    if data_loaded:
        print('Translating buss list into connection list...')
        connections = pro.get_connections_from_buss_dict(busses_dict)
        print('Done')
        print('Creating DOT file...')
        dot = Digraph() #Use property engine=<layout_name> for different layout engines
        dot.attr('graph', splines='ortho')
        dot = add_modules_to_dot(dot, module_set)
        dot = add_connections_to_dot(dot, connections)
        print('Done')

        output_file_name = remove_extension_from_name(data_name)
        dot.save('{}.gv'.format(output_file_name))
        os.system('/cad/gnu/anaconda/3-2.4.0/bin/fdp -Goverlap=scale -Tpdf {0}.gv -o {0}.pdf'.format(output_file_name))
        print('Diagram saved as {0}.pdf and {0}.gv'.format(output_file_name))

        if y_n_input("Do you want to view the file? [y/n] >> "):
            os.system('okular {}.pdf'.format(output_file_name))
        return
    else:
        print('Error: No data have been loaded.')
    return


# Generate a diagram detailing masters and slaves connected to the masters
def generate_master_diagram():
    print('Creating DOT file...')
    dot = Digraph()
    dot.attr('graph', splines='ortho')
    dot.attr('node', shape='box')
    dot.attr('edge', penwidth='3') 

    for m in masters:
        reduced_master_name = reduce_hierarchical_path_name(masters[m].name)
        if enable_reduced_names:
            dot.node(masters[m].name, label=reduced_master_name)
        else:
            dot.node(masters[m].name, label=masters[m].name)
        #dot.node(masters[m].name)
        for s in masters[m].slaves:
            reduced_name = reduce_hierarchical_path_name(s)
            if enable_reduced_names:
                dot.node(s, label=reduced_name)
            else:
                dot.node(s, label=s)
            #dot.node(s)
            dot.edge(masters[m].name, s)

    output_file_name = remove_extension_from_name(data_name)
    dot.save('{}.gv'.format(output_file_name))
    print('Done')

    print('Creating PDF file...')
    os.system('/cad/gnu/anaconda/3-2.4.0/bin/fdp -Goverlap=scale -Tpdf {0}.gv -o {0}.pdf'.format(output_file_name))
    print('Done')
    
    if y_n_input("Do you want to view the file? [y/n] >> "):
        os.system('okular {}.pdf'.format(output_file_name))
    return


# Toggles whether the name reduction is applied or not.
def toggle_name_reduction():
    global enable_reduced_names
    toggle = False
    if enable_reduced_names:
        print('Name reduction is currently ENABLED')
        toggle = y_n_input('Disable reduction of names? [y/n] >> ')
    else:
        print('Name reduction is currently DISABLED')
        toggle = y_n_input('Enable reduction of names? [y/n] >> ')

    if toggle:
        enable_reduced_names = not enable_reduced_names


# Checks the command line agruments and takes correspondingly actions
def check_command_line_args():
    global enable_reduced_names
    global enable_debug_mode
    import_file_path = ""
    import_at_init = False

    try:
        opts, args = getopt.getopt(sys.argv[1:],"ni:d")
    except:
        print('Error: Invalid arguments.')
        return

    for opt, arg in opts:
        if opt == '-i':
            import_at_init = True
            import_file_path = arg
        elif opt == '-n':
            enable_reduced_names = False
        elif opt == '-d':
            enable_debug_mode = True

    if import_at_init:
        import_file(import_file_path)
    return 


# Print the set of modules obtained from data
def list_modules():
    if data_loaded:
        print('----------------------------MODULES---------------------------------')
        for mod in module_set:
            if enable_reduced_names:
                print(reduce_hierarchical_path_name(mod))
            else: 
                print(mod)
        print('--------------------------------------------------------------------')
    else:
        print('Error: No data have been loaded.')
    return 


#Print the dictionary of busses obtained from data
def list_connections():
    if data_loaded:
        print('----------------------------CONNECTIONS-----------------------------')
        for bus in busses_dict:
            nb = pro.Bus(busses_dict[bus].get_value())
            if enable_reduced_names:
                for r in busses_dict[bus].read_sources:
                    nb.add_read_source(reduce_hierarchical_path_name(r))
                for n in busses_dict[bus].names:
                    nb.add_name(reduce_hierarchical_path_name(n))
                nb.set_write_source(reduce_hierarchical_path_name(busses_dict[bus].get_write_source()))
            else:
                nb = busses_dict[bus]

            if enable_debug_mode:
                print(nb)
                print("")
            else:
                print('Name(s) of Connection: {}\n\tWrite Source: {}\n\tRead Source(s): {}\n'.format(
                    nb.names, nb.get_write_source(), nb.read_sources
                ))
        print('--------------------------------------------------------------------')
    else:
        print('Error: No data have been loaded.')
    return

#Print the dictionary of master modules, and what slaves is connected to them.
def list_masters():
    if data_loaded:
        print('----------------------------MASTERS---------------------------------') 
        for m in masters.values():
            new_m = m
            if enable_reduced_names:
                new_m.name = reduce_hierarchical_path_name(new_m.name)
                new_slaves = set()
                for s in new_m.slaves:
                    new_slaves.add(reduce_hierarchical_path_name(s))
                new_m.slaves = new_slaves

            if enable_debug_mode:
                print(new_m)
            else:
                #print(m.slaves)
                print('Master: {}\nSlaves: {}\n'.format(new_m.name, new_m.slaves))
            #print(masters[m])
        print('--------------------------------------------------------------------')
    else:
        print('Error: No data have been loaded.')
    return

# Imports a specified textfile
def import_file(filename):
    global data_name
    global data_loaded
    global module_set
    global busses_dict
    global masters

    print('Parsing {}...'.format(filename))
    data_points = bip.parse_text_file(filename)
    if data_points == -1:
        print('Import failed')
        return
    print('Done')

    print('Obtaining list of modules...')
    module_set = pro.get_modules_from_data_points(data_points)
    print('Done')

    print('Obtaining list of busses...')
    busses_dict = pro.get_busses_from_data_points(data_points)
    print('Done')

    print('Obtaining list of masters and their slaves...')
    masters = pro.get_master_list(data_points)
    print('Done')

    data_name = filename
    data_loaded = True 
    print('Processing finished. Imported {} modules.'.format(len(module_set)))
    return


# Prints a list of all available commands to the user
def get_commands():
    print("""----------------------------COMMANDS--------------------------------
    import <filename.txt>\tImport meta-information from specified textfile.
    listmod\t\t\tDisplay a list of all modules from extracted data.
    listcon\t\t\tDisplay a list of all connection between modules.
    listmas\t\t\tDisplay a list of all masters, and slaves connected to the master.
    genmas\t\t\tGenerate a pdf image of the relationship between masters and slave.
    gencon\t\t\tGenerate a diagram of the connection between modules in the system.
    togname\t\t\tToggle whether name reduction is enabled or not.
    help\t\t\tShows this list of available commands.
    quit\t\t\tQuits the tool. (Can also use 'exit'.)
----------------------------ARGUMENTS-------------------------------
    -i <filename.txt>\t\tImport file at initialization of the tool
    -n\t\t\t\tTurn off name reduction at startup
    -d\t\t\t\tStart tool in debug mode
--------------------------------------------------------------------""")
    return


# main loop
print('Processing tool for extracted bus information')
print('---------------------------------------------')
check_command_line_args()
print('Data loaded: {}'.format(data_name))
print('Reduced Names: {}'.format(enable_reduced_names))
if enable_debug_mode:
    print('DEBUG MODE')
print('---------------------------------------------')

while True:
    command = input('>> ').lower()
    if command == 'help':
        get_commands()
    elif command == 'quit' or command == 'exit':
        sys.exit(0)
    elif command == 'listmod':
        list_modules()
    elif command == 'listcon':
        list_connections()
    elif command == 'listmas':
        list_masters()
    elif command == 'genmas':
        generate_master_diagram()
    elif command == 'gencon':
        generate_connection_diagram()
    elif command == 'togname':
        toggle_name_reduction()
    elif command.split(' ')[0] == 'import':
        import_file(command.split(' ')[1]) 
    else:
        print('Invalid command. Type "help" for list of available commands.')
